@@ -1,16 +1,16 @@
 /*
  * Do rate lookups and set destination route and maximum call length based on that
  * 
- * Copyright (C) 2003 by Troll Phone Networks AS
+ * Copyright (C) 2003, 2004, 2005 by Voop as
  *
  * This program is distributed under the terms of the GNU General Public License
  * as published by the Free Software Foundation; either version 2, or (at your
  * option) any later version.
  *
- *	$Id: rate_engine.c,v 1.26 2004/08/26 16:50:18 tholo Exp $
+ *	$Id: rate_engine.c,v 1.36 2005/11/03 22:00:59 tholo Exp $
  */
 
-#ifdef __linux__
+#if defined(__linux__) || defined(__GLIBC__)
 #define _GNU_SOURCE
 #endif
 
@@ -20,7 +20,7 @@
 
 #include <pthread.h>
 
-#include <mysql/mysql.h>
+#include <mysql.h>
 
 #include <pcre.h>
 #include "pcre_subst.h"
@@ -35,24 +35,25 @@
 #include <asterisk/cli.h>
 #include <asterisk/manager.h>
 #include <asterisk/utils.h>
+#include <asterisk/term.h>
 
 static char *tdesc = "Call Routing and Rating Application";
 static char *cdr_name = "ratecall";
 static char *routecall_app = "RouteCall";
 static char *routecall_synopsis = "Do cost based routing of a call and set call time limit";
 static char *routecall_descrip =
-"  RouteCall(extension[|flags]): Does Least Cost Routing of a call, and\n"
-"sets  an  absolute timeout  on  call length  based on customers credit\n"
+"  RouteCall(extension[|timeout][|flags][|URL]):  Do Least Cost Routing\n"
+"of a call,  and set an absolute timeout on call length based on credit\n"
 "worthiness.  If there exists a priority n + 101, and no route could be\n"
-"found,  then the  channel  will be set up to continue at that priority\n"
+"found,  then  the channel  will be set up to continue at that priority\n"
 "level.\n"
-"  If a route  is  found, a dialstring suitable  for  use with the Dial\n"
-"application is created in the variable ${DESTINATIONnn}, to be used as\n"
-"Dial(${DESTINATIONnn}) where 'nn' is a number  from 1 and up, with the\n"
-"route in ${DESTINATION1} being the lowest-cost route found,  the route\n"
-"in  ${DESTINATION2} being  the  next  lowest rate found and so on.  An\n"
-"AbsoluteTimeout limit may also be set for the call.\n"
-"  Always returns 0.\n";
+"  If a route is found, the  Dial  application is called for the egress\n"
+"route with the optional timeout, flags and URL arguments. There may be\n"
+"an AbsoluteTimeout limit set for the call.\n"
+"  This application returns -1 if the originating channel hangs up,  or\n"
+"if the call is  bridged  and  either  of the  parties  in  the  bridge\n"
+"terminate the call.\n"
+"  For possible flags, see the Dial application.\n";
 
 STANDARD_LOCAL_USER;
 
@@ -178,10 +179,9 @@ static int load_rates(int);
  */
 static int routecall_exec(struct ast_channel *chan, void *data)
 {
-    char destvar[20], firstvar[20], restvar[20], startcost[20], periodcost[20];
-    char trialcost[20];
-    char *args, *number, *flags, *destination, *sub;
-    int res = 0, maxcall = 0, i, cmpval;
+    char *dialer, *args, *number, *timeout, *flags, *url, *destination, *sub, *status;
+    int res = 0, res2, maxcall = 0, i, cmpval, len, pri;
+    struct ast_app *dial;
     struct localuser *u;
     char tempstr[16];
     splayval_t *val;
@@ -201,26 +201,49 @@ static int routecall_exec(struct ast_channel *chan, void *data)
      * Verify that arguments were supplied
      */
     if (data == NULL || *(char *)data == '\0') {
-	ast_log(LOG_WARNING, "RouteCall requires an argument (number[|flags])\n");
+	ast_log(LOG_WARNING, "RouteCall requires an argument (number[|timeout][|flags][|URL])\n");
 	return -1;
     }
 
     LOCAL_USER_ADD(u);
 
     /*
-     * Parse out arguments to number and (optional) flags
+     * Parse out arguments to number, (optional) timeout, (optional) flags and (optional) URL
      */
     args = ast_strdupa((char *)data);
     number = strsep(&args, "|");
-    flags = strsep(&args, "\0");
+    if (number) {
+	timeout = strsep(&args, "|");
+	if (timeout) {
+	    strtol(timeout, &flags, 10);
+	    if (flags) {
+		flags = timeout;
+		timeout = NULL;
+	    }
+	}
+    }
+    else 
+	timeout = NULL;
+    if (timeout)
+	flags = strsep(&args, "|");
+    else
+	flags = NULL;
+    if (flags)
+	url = strsep(&args, "\0");
+    else
+	url = NULL;
 
     /*
      * Make empty strings be NULL pointers instead
      */
     if (number && *number == '\0')
 	number = NULL;
+    if (timeout && *timeout == '\0')
+	timeout = NULL;
     if (flags && *flags == '\0')
 	flags = NULL;
+    if (url && *url == '\0')
+	url = NULL;
 
     /*
      * We can't do any routing if we don't have any rates or
@@ -302,11 +325,22 @@ static int routecall_exec(struct ast_channel *chan, void *data)
      */
     time(&now);
 
+    /*
+     * XXX -- copy the rates we found
+     */
+
+    /*
+     * Release the lock
+     */
+    ast_mutex_unlock(&ratelock);
+
     /*
      * The in-memory rates are already sorted from least to
      * most expensive
      */
-    for (i = 0, val = s->value; val; val = val->next) {
+    for (pri = chan->priority, i = 0, val = s->value, res = -1, res2 = 0;
+	 val;
+	 val = val->next, chan->priority = pri) {
 	/*
 	 * If we do not have the actual egress route, skip the rate
 	 */
@@ -325,45 +359,113 @@ static int routecall_exec(struct ast_channel *chan, void *data)
 	     val->rate->expires_date > now))
 	    continue;
 
-	snprintf(destvar, sizeof(destvar), "DESTINATION%d", i + 1);
-	snprintf(firstvar, sizeof(firstvar), "FIRSTLEN%d", i + 1);
-	snprintf(restvar, sizeof(restvar), "RESTLEN%d", i + 1);
-	snprintf(startcost, sizeof(startcost), "STARTCOST%d", i + 1);
-	snprintf(periodcost, sizeof(periodcost), "PERIODCOST%d", i + 1);
-	snprintf(trialcost, sizeof(trialcost), "TRIALCOST%d", i + 1);
+	/*
+	 * Signal that at least one possible egress route was
+	 * found
+	 */
+	res = 0;
 
-	sub = malloc(strlen(val->rate->egress->technology) +
-		     strlen(val->rate->egress->peer) +
-		     strlen(val->rate->egress->substitute) +
-		     3);	/* Two /'s and a NUL */
+	/*
+	 * Build substitution pattern
+	 */
+	len = strlen(val->rate->egress->technology) +
+	      strlen(val->rate->egress->peer) +
+	      strlen(val->rate->egress->substitute) +
+	      3;	/* Two /'s and a NUL */
+	if (timeout)
+	    len += strlen(timeout) + 1;
+	if (flags)
+	    len += strlen(flags) + 1;
+	if (url)
+	    len += strlen(url) + 1;
+	sub = malloc(len);
 	strcpy(sub, val->rate->egress->technology);
 	strcat(sub, "/");
 	strcat(sub, val->rate->egress->peer);
 	strcat(sub, "/");
 	strcat(sub, val->rate->egress->substitute);
+	if (timeout) {
+	    strcat(sub, "|");
+	    strcat(sub, timeout);
+	}
+	if (flags) {
+	    strcat(sub, "|");
+	    strcat(sub, flags);
+	}
+	if (url) {
+	    strcat(sub, "|");
+	    strcat(sub, url);
+	}
 	destination = pcre_subst(val->rate->egress->pcre, val->rate->egress->extra, number, strlen(number), 0, 0, sub);
 	if (sub)
 	    free(sub);
 
-	res = 0;
-	pbx_builtin_setvar_helper(chan, destvar, destination);
+	/*
+	 * Set an AbsoluteTimeout on the call, based on customer
+	 * credit worthiness and call cost for the destination
+	 * we have found a route for
+	 */
+	ast_channel_setwhentohangup(chan, maxcall);
+	if (option_verbose > 3)
+	    ast_verbose( VERBOSE_PREFIX_4 "Set AbsoluteTimeout to %d\n", maxcall);
+
+	pbx_builtin_setvar_helper(chan, "DESTINATION", destination);
 	snprintf(tempstr, sizeof(tempstr), "%d", val->rate->firstperiod);
-	pbx_builtin_setvar_helper(chan, firstvar, tempstr);
+	pbx_builtin_setvar_helper(chan, "FIRSTLEN", tempstr);
 	snprintf(tempstr, sizeof(tempstr), "%d", val->rate->periods);
-	pbx_builtin_setvar_helper(chan, restvar, tempstr);
+	pbx_builtin_setvar_helper(chan, "RESTLEN", tempstr);
 	snprintf(tempstr, sizeof(tempstr), "%d", val->rate->startcost);
-	pbx_builtin_setvar_helper(chan, startcost, tempstr);
+	pbx_builtin_setvar_helper(chan, "STARTCOST", tempstr);
 	snprintf(tempstr, sizeof(tempstr), "%d", val->rate->periodcost);
-	pbx_builtin_setvar_helper(chan, periodcost, tempstr);
+	pbx_builtin_setvar_helper(chan, "PERIODCOST", tempstr);
 	snprintf(tempstr, sizeof(tempstr), "%d", val->rate->trialcost);
-	pbx_builtin_setvar_helper(chan, trialcost, tempstr);
-	if (option_verbose > 3) {
-	    if (i)
-		ast_verbose( VERBOSE_PREFIX_4 "Alternate route as %s\n", destination);
-	    else
-		ast_verbose( VERBOSE_PREFIX_4 "Route as %s\n", destination);
+	pbx_builtin_setvar_helper(chan, "TRIALCOST", tempstr);
+
+	/*
+	 * Attempt to place a call using the selected egress route
+	 */
+	dialer = pbx_builtin_getvar_helper(chan, "DIALROUTE");
+	if (dialer == NULL)
+	    dialer = "Dial";
+	dial = pbx_findapp(dialer);
+	if (dial) {
+	    if (option_debug)
+		ast_log(LOG_DEBUG, "Launching '%s'\n", dialer);
+	    if (option_verbose > 2) {
+		char tmp[80], tmp2[80], tmp3[80];
+
+		ast_verbose( VERBOSE_PREFIX_3 "Executing %s(\"%s\", \"%s\") in new stack (route %d)\n", 
+			     term_color(tmp, dialer, COLOR_BRCYAN, 0, sizeof(tmp)),
+			     term_color(tmp2, chan->name, COLOR_BRMAGENTA, 0, sizeof(tmp2)),
+			     term_color(tmp3, destination, COLOR_BRMAGENTA, 0, sizeof(tmp3)),
+			     i + 1);
+	    }
+
+	    res2 = pbx_exec(chan, dial, destination, 1);
+	    if (option_verbose > 3)
+		ast_verbose( VERBOSE_PREFIX_4 "Dial returned %d\n", res2);
+
+	    /*
+	     * Reset AbsoluteTimeout now that the call has been
+	     * completed
+	     */
+	    ast_channel_setwhentohangup(chan, 0);
+
+	    /*
+	     * If the channel was unavailable we need to try the next provider,
+	     * if any are provided.  Otherwise we are all done
+	     */
+	    status = pbx_builtin_getvar_helper(chan, "DIALSTATUS");
+	    if (strcasecmp(dialer, "NoOp") != 0 &&
+		(status == 0 || strcasecmp(status, "CHANUNAVAIL") != 0)) {
+		i = 1;
+		break;
+	    }
 	}
+	else
+	    ast_log(LOG_WARNING, "No dial application registered\n");
 	pcre_free(destination);
+
 	i++;
     }
 
@@ -374,14 +476,10 @@ static int routecall_exec(struct ast_channel *chan, void *data)
      */
     if (i == 0) {
 	ast_log(LOG_WARNING, "No valid egress routes for prefix '%s'\n", s->key);
-	goto errout;
+	chan->priority = pri;
+	goto errout2;
     }
 
-    /*
-     * Release the lock
-     */
-    ast_mutex_unlock(&ratelock);
-
     /*
      * If we could not find a match for the number, we should
      * log an entry to the error table so that there will be
@@ -391,25 +489,25 @@ static int routecall_exec(struct ast_channel *chan, void *data)
     if (res < 0) {
 	res = 0;
 	ast_log(LOG_WARNING, "No route could be made for phone number %s\n", number);
-	goto errout;
+	chan->priority = pri;
+	goto errout2;
     }
 
     /*
-     * Set an AbsoluteTimeout on the call, based on customer
-     * credit worthiness and call cost for the destination
-     * we have found a route for
+     * Make sure we return the status from the Dial command so that
+     * the dialplan will behave as expected
      */
-    ast_channel_setwhentohangup(chan, maxcall);
-    if (option_verbose > 3)
-	ast_verbose( VERBOSE_PREFIX_4 "Set AbsoluteTimeout to %d\n", maxcall);
+    res = res2;
 
     goto out;
+
 errout:
     /*
      * Release the lock
      */
     ast_mutex_unlock(&ratelock);
 
+errout2:
     /*
      * Jump to priority + 101, since we got an error
      */
@@ -436,11 +534,9 @@ out:
 static int cdr_ratecall(struct ast_cdr *cdr)
 {
     char *ani, *src, *dst, *dcontext, *channel, *dstchannel, *lastapp, *lastdata;
-    char starttime[32], answertime[32], endtime[32], destvar[32], firstvar[32];
-    char trialvar[32];
-    int len, i, first = 0, start = 0, each = 0, rest = 0, trial = 0;
+    char starttime[32], answertime[32], endtime[32];
+    int len, first = 0, start = 0, each = 0, rest = 0, trial = 0;
     char *account, *uniqueid, *destination = "", *end;
-    char restvar[32], startvar[32], periodvar[32];
     struct ast_channel *chan;
     long long cost = 0;
     cdr_queue_t *q;
@@ -467,30 +563,18 @@ static int cdr_ratecall(struct ast_cdr *cdr)
 	chan = ast_channel_walk_locked(chan);
     }
     if (chan) {
-	for (i = 1;; i++) {
-	    snprintf(destvar, sizeof(destvar), "DESTINATION%d", i);
-	    snprintf(firstvar, sizeof(firstvar), "FIRSTLEN%d", i);
-	    snprintf(restvar, sizeof(restvar), "RESTLEN%d", i);
-	    snprintf(startvar, sizeof(startvar), "STARTCOST%d", i);
-	    snprintf(periodvar, sizeof(periodvar), "PERIODCOST%d", i);
-	    snprintf(trialvar, sizeof(trialvar), "TRIALCOST%d", i);
-
-	    dst = pbx_builtin_getvar_helper(chan, destvar);
-	    if (dst == NULL)
-		break;
-	    if (strcmp(dst, cdr->lastdata) == 0) {
-		dst = pbx_builtin_getvar_helper(chan, firstvar);
-		first = strtoul(dst, &end, 10);
-		dst = pbx_builtin_getvar_helper(chan, restvar);
-		rest = strtoul(dst, &end, 10);
-		dst = pbx_builtin_getvar_helper(chan, startvar);
-		start = strtoul(dst, &end, 10);
-		dst = pbx_builtin_getvar_helper(chan, periodvar);
-		each = strtoul(dst, &end, 10);
-		dst = pbx_builtin_getvar_helper(chan, trialvar);
-		trial = strtoul(dst, &end, 10);
-		break;
-	    }
+	dst = pbx_builtin_getvar_helper(chan, "DESTINATION");
+	if (dst && strcmp(dst, cdr->lastdata) == 0) {
+	    dst = pbx_builtin_getvar_helper(chan, "FIRSTLEN");
+	    first = strtoul(dst, &end, 10);
+	    dst = pbx_builtin_getvar_helper(chan, "RESTLEN");
+	    rest = strtoul(dst, &end, 10);
+	    dst = pbx_builtin_getvar_helper(chan, "STARTCOST");
+	    start = strtoul(dst, &end, 10);
+	    dst = pbx_builtin_getvar_helper(chan, "PERIODCOST");
+	    each = strtoul(dst, &end, 10);
+	    dst = pbx_builtin_getvar_helper(chan, "TRIALCOST");
+	    trial = strtoul(dst, &end, 10);
 	}
 	ast_mutex_unlock(&chan->lock);
     }
@@ -1498,6 +1582,20 @@ static void clear_rates(void)
     rates = NULL;
 }
 
+static int rate_compare(const rate_t *rate1, const rate_t *rate2)
+{
+    long cost1, cost2;
+
+    cost1 = rate1->startcost + ((avg_call_length - rate1->firstperiod <= 0 ? 0 : avg_call_length - rate1->firstperiod) + rate1->periods - 1) / rate1->periods * rate1->periodcost;
+    cost2 = rate2->startcost + ((avg_call_length - rate2->firstperiod <= 0 ? 0 : avg_call_length - rate2->firstperiod) + rate2->periods - 1) / rate2->periods * rate2->periodcost;
+
+    if (cost1 < cost2)
+	return -1;
+    else if (cost1 > cost2)
+	return 1;
+    return 0;
+}
+
 /*
  * Inject a rate entry into the in-memory table -- when adding
  * a rate to an existing prefix, we should have a linked list
@@ -1505,7 +1603,7 @@ static void clear_rates(void)
  */
 static void inject_rate(rate_t *rate)
 {
-    splayval_t *val;
+    splayval_t *val, *newval;
     char id[12];
     splay_t *s;
 
@@ -1519,18 +1617,42 @@ static void inject_rate(rate_t *rate)
 	rate->egress = val->egress;
     }
 
-    val = malloc(sizeof(splayval_t));
-    val->egress = NULL;
-    val->rate = rate;
+    newval = malloc(sizeof(splayval_t));
+    newval->egress = NULL;
+    newval->rate = rate;
 
     s = search_splay(rates, rate->prefix);
     if (s) {
-	val->next = s->value;
-	s->value = val;
+	val = s->value;
+	if (rate_compare(newval->rate, val->rate) < 0) {
+	    newval->next = val;
+	    s->value = newval;
+	    newval = NULL;
+	}
+	else {
+	    for (val = s->value; val->next ; val = val->next) {
+		if (rate_compare(newval->rate, val->next->rate) < 0) {
+		    newval->next = val->next;
+		    val->next = newval;
+		    newval = NULL;
+		    break;
+		}
+		else if (val->next == NULL) {
+		    val->next = newval;
+		    newval->next = NULL;
+		    newval = NULL;
+		    break;
+		}
+	    }
+	    if (val && newval) {
+		val->next = newval;
+		newval->next = NULL;
+	    }
+	}
     }
     else {
-	val->next = NULL;
-	insert_splay(rates, rate->prefix, val);
+	newval->next = NULL;
+	insert_splay(rates, rate->prefix, newval);
     }
 }
 
